
/* Uses the /dev/mem addresses to change GPIO lines. 
    Based on code seen in 

    http://elinux.org/RPi_Low-level_peripherals

    We can access gpios in device drivers as well, but then we have
    to re-do the kernel; we may not always want to have to do that.
*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <unistd.h>
#include <stdint.h>

/* #define BCM2708_PERI_BASE 0x20000000 Pi A, A+, B, B+ 256MB, 512 MB */ 
/* #define BCM2708_PERI_BASE 0x3F000000 Pi 2 1GB */ 
/* Somehow this translates to 0x7E00 0000  */ 

/* #define GPIO_BASE (BCM2708_PERI_BASE+0x200000) */ 
/* Makes mapping to 0x7E20 0000 */

#define BLOCK_SIZE (4*1024)
#define PAGE_SIZE (4*1024)

// GPIO setup macros. Always use INP_GPIO(x) before using OUT_GPIO(x) or SET_GPIO_ALT(x,y)
#define INP_GPIO(g) *(gpio+((g)/10)) &= ~(7<<(((g)%10)*3))
#define OUT_GPIO(g) *(gpio+((g)/10)) |=  (1<<(((g)%10)*3))

/* Access bits in the 3-bit groups at +00 +04 +08 +0C +10 +14 
    Values for each of these triads:
    000 = output
    001 = input
    010 = Alternate function 5
    011 = Alternate function 4
    100 = Alternate function 0
    101 = Alternate function 1
    110 = Alternate function 2
    111 = Alternate function 3
*/

struct tag_gpioinfo
{
    int gpiono;
    char *pinnoAB2;      /* Pin-number on the Model A, Model B v2  (26-pin and 8-pin headers) */
    char *pinnoBplus;    /* Pin-number on Model A+, B+, Pi2 (40-pin header) */
    // char *pinnoCOMO;  /* Compute-module pin no as and when available */
    char *fx[8];         /* Function descriptions */
    int GPFSELindex;     /* (offset for the GPFSEL). */
    uint32_t GPFSELmask; /* mask in GPFSEL. */
    int ONEoffset;       /* Offset from the XXX0 register */
    uint32_t ONEmask;    /* mask for the 1-bit registers. */
};


/* Available gpios on the headers: */ 
struct tag_gpioinfo thegpinfo[] =  {
{ 0, "NA-00", "J8-27", {"input", "output", "F5=.DU",     "F4=.DU",       "F0=SDA0",      "F1=SA5",        "F2=.RE",     "F3=.DU"           },0, 00000000007, 0, 0x00000001 },  /* 0 */
{ 1, "NA-00", "J8-28", {"input", "output", "F5=.DU",     "F4=.DU",       "F0=SCL0",      "F1=SA4",        "F2=.RE",     "F3=.DU"           },0, 00000000070, 0, 0x00000002 },  /* 1 */
{ 2, "P1-03", "J8-03", {"input", "output", "F5=.DU",     "F4=.DU",       "F0=SDA1",      "F1=SA3",        "F2=.RE",     "F3=.DU"           },0, 00000000700, 0, 0x00000004 },  /* 2 */
{ 3, "P1-05", "J8-05", {"input", "output", "F5=.DU",     "F4=.DU",       "F0=SCL1",      "F1=SA2",        "F2=.RE",     "F3=.DU"           },0, 00000007000, 0, 0x00000008 },  /* 3 */
{ 4, "P1-07", "J8-07", {"input", "output", "F5=ARM_TDI", "F4=.DU",       "F0=GPCLK0",    "F1=SA1",        "F2=.RE",     "F3=.DU"           },0, 00000070000, 0, 0x00000010 },  /* 4 */
{ 5, "NA-00", "J8-29", {"input", "output", "F5=ARM_TDO", "F4=.DU",       "F0=GPCLK1",    "F1=SA0",        "F2=.RE",     "F3=.DU"           },0, 00000700000, 0, 0x00000020 },  /* 5 */
{ 6, "NA-00", "J8-31", {"input", "output", "F5=ARM_RTCK","F4=.DU",       "F0=GPCLK2",    "F1=SOE_N/SE",   "F2=.RE",     "F3=.DU"           },0, 00007000000, 0, 0x00000040 },  /* 6 */
{ 7, "P1-26", "J8-26", {"input", "output", "F5=.DU",     "F4=.DU",       "F0=SPI0_CE1_N","F1=SWE_N/SRW_N","F2=.RE",     "F3=.DU"           },0, 00070000000, 0, 0x00000080 },  /* 7 */
{ 8, "P1-24", "J8-24", {"input", "output", "F5=.DU",     "F4=.DU",       "F0=SPI0_CE0_N","F1=SD0",        "F2=.RE",     "F3=.DU"           },0, 00700000000, 0, 0x00000100 },  /* 8 */
{ 9, "P1-21", "J8-21", {"input", "output", "F5=.DU",     "F4=.DU",       "F0=SPI0_MISO", "F1=SD1",        "F2=.RE",     "F3=.DU"           },0, 07000000000, 0, 0x00000200 },  /* 9 */
{10, "P1-19", "J8-19", {"input", "output", "F5=.DU",     "F4=.DU",       "F0=SPI0_MOSI", "F1=SD2",        "F2=.RE",     "F3=.DU"           },1, 00000000007, 0, 0x00000400 },  /* 10 */
{11, "P1-23", "J8-23", {"input", "output", "F5=.DU",     "F4=.DU",       "F0=SPI0_SCLK", "F1=SD3",        "F2=.RE",     "F3=.DU"           },1, 00000000070, 0, 0x00000800 },  /* 11 */
{12, "NA-00", "J8-32", {"input", "output", "F5=ARM_TMS", "F4=.DU",       "F0=PWM0",      "F1=SD4",        "F2=.RE",     "F3=.DU"           },1, 00000000700, 0, 0x00001000 },  /* 12 */
{13, "NA-00", "J8-33", {"input", "output", "F5=ARM_TCK", "F4=.DU",       "F0=PWM1",      "F1=SD5",        "F2=.RE",     "F3=.DU"           },1, 00000007000, 0, 0x00002000 },  /* 13 */
{14, "P1-08", "J8-08", {"input", "output", "F5=TXD1",    "F4=.DU",       "F0=TXD0",      "F1=SD6",        "F2=.RE",     "F3=.DU"           },1, 00000070000, 0, 0x00004000 },  /* 14 */
{15, "P1-10", "J8-10", {"input", "output", "F5=RXD1",    "F4=.DU",       "F0=RXD0",      "F1=SD7",        "F2=.RE",     "F3=.DU"           },1, 00000700000, 0, 0x00008000 },  /* 15 */
{16, "NA-00", "J8-36", {"input", "output", "F5=CTS1",    "F4=SPI1_CE2_N","F0=.RE",       "F1=SD8",        "F2=.RE",     "F3=CTS0"          },1, 00007000000, 0, 0x00010000 },  /* 16 */
{17, "P1-11", "J8-11", {"input", "output", "F5=RTS1",    "F4=SPI1_CE1_N","F0=.RE",       "F1=SD9",        "F2=.RE",     "F3=RTS0"          },1, 00070000000, 0, 0x00020000 },  /* 17 */
{18, "P1-12", "J8-12", {"input", "output", "F5=PWM0",    "F4=SPI1_CE0_N","F0=PCM_CLK",   "F1=SD10",       "F2=.RE",     "F3=BSCSL SDA/MOSI"},1, 00700000000, 0, 0x00040000 },  /* 18 */
{19, "NA-00", "J8-35", {"input", "output", "F5=PWM1",    "F4=SPI1_MISO", "F0=PCM_FS",    "F1=SD11",       "F2=.RE",     "F3=BSCSL SCL/SCLK"},1, 07000000000, 0, 0x00080000 },  /* 19 */
{20, "NA-00", "J8-38", {"input", "output", "F5=GPCLK0",  "F4=SPI1_MOSI", "F0=PCM_DIN",   "F1=SD12",       "F2=.RE",     "F3=BSCSL -/MISO"  },2, 00000000007, 0, 0x00100000 },  /* 20 */
{21, "NA-00", "J8-40", {"input", "output", "F5=GPCLK1",  "F4=SPI1_SCLK", "F0=PCM_DOUT",  "F1=SD13",       "F2=.RE",     "F3=BSCSL -/CE_N"  },2, 00000000070, 0, 0x00200000 },  /* 21 */
{22, "P1-15", "J8-15", {"input", "output", "F5=.DU",     "F4=ARM_TRST",  "F0=.RE",       "F1=SD14",       "F2=.RE",     "F3=SD1_CLK"       },2, 00000000700, 0, 0x00400000 },  /* 22 */
{23, "P1-16", "J8-16", {"input", "output", "F5=.DU",     "F4=ARM_RTCK",  "F0=.RE",       "F1=SD15",       "F2=.RE",     "F3=SD1_CMD"       },2, 00000007000, 0, 0x00800000 },  /* 23 */
{24, "P1-18", "J8-18", {"input", "output", "F5=.DU",     "F4=ARM_TDO",   "F0=.RE",       "F1=SD16",       "F2=.RE",     "F3=SD1_DAT0"      },2, 00000070000, 0, 0x01000000 },  /* 24 */
{25, "P1-22", "J8-22", {"input", "output", "F5=.DU",     "F4=ARM_TCK",   "F0=.RE",       "F1=SD17",       "F2=.RE",     "F3=SD1_DAT1"      },2, 00000700000, 0, 0x02000000 },  /* 25 */
{26, "NA-00", "J8-37", {"input", "output", "F5=.DU",     "F4=ARM_TDI",   "F0=.RE",       "F1=.RE",        "F2=.RE",     "F3=SD1_DAT2"      },2, 00007000000, 0, 0x04000000 },  /* 26 */
{27, "P1-13", "J8-13", {"input", "output", "F5=.DU",     "F4=ARM_TMS",   "F0=.RE",       "F1=.RE",        "F2=.RE",     "F3=SDI_DAT3"      },2, 00070000000, 0, 0x08000000 },  /* 27 */
{28, "P5-03", "NA-00", {"input", "output", "F5=.DU",     "F4=.DU",       "F0=SDA0",      "F1=SA5",        "F2=PCM_CLK", "F3=.RE"           },2, 00700000000, 0, 0x10000000 },  /* 28 */
{29, "P5-04", "NA-00", {"input", "output", "F5=.DU",     "F4=.DU",       "F0=SCL0",      "F1=SA4",        "F2=PCM_FS",  "F3=.RE"           },2, 07000000000, 0, 0x20000000 },  /* 29 */
{30, "P5-05", "NA-00", {"input", "output", "F5=CTS1",    "F4=.DU",       "F0=.RE",       "F1=SA3",        "F2=PCM_DIN", "F3=CTS0"          },3, 00000000007, 0, 0x40000000 },  /* 30 */
{31, "P5-06", "NA-00", {"input", "output", "F5=RTS1",    "F4=.DU",       "F0=.RE",       "F1=SA2",        "F2=PCM_DOUT","F3=RTS0"          },3, 00000000070, 0, 0x80000000 },  /* 31 */
{32, "NA-00", "NA-00", {"input", "output", "F5=TXD1",    "F4=.DU",       "F0=GPCLK0",    "F1=SA1",        "F2=.RE",     "F3=TXD0"          },3, 00000000700, 1, 0x00000001 },  /* 32 */
{33, "NA-00", "NA-00", {"input", "output", "F5=RXD1",    "F4=.DU",       "F0=.RE",       "F1=SA0",        "F2=.RE",     "F3=RXD0"          },3, 00000007000, 1, 0x00000002 },  /* 33 */
{34, "NA-00", "NA-00", {"input", "output", "F5=.DU",     "F4=.DU",       "F0=GPCLK0",    "F1=SOE_N/SE",   "F2=.RE",     "F3=.RE"           },3, 00000070000, 1, 0x00000004 },  /* 34 */
{35, "NA-00", "NA-00", {"input", "output", "F5=.DU",     "F4=.DU",       "F0=SPI0_CE1_N","F1=SWE_N/SRW_N","F2=.DU",     "F3=.RE"           },3, 00000700000, 1, 0x00000008 },  /* 35 */
{36, "NA-00", "NA-00", {"input", "output", "F5=.DU",     "F4=.DU",       "F0=SPI0_CE0_N","F1=SD0",        "F2=TXD0",    "F3=.RE"           },3, 00007000000, 1, 0x00000010 },  /* 36 */
{37, "NA-00", "NA-00", {"input", "output", "F5=.DU",     "F4=.DU",       "F0=SPI0_MISO", "F1=SD1",        "F2=RXD0",    "F3=.RE"           },3, 00070000000, 1, 0x00000020 },  /* 37 */
{38, "NA-00", "NA-00", {"input", "output", "F5=.DU",     "F4=.DU",       "F0=SPI0_MOSI", "F1=SD2",        "F2=RTS0",    "F3=.RE"           },3, 00700000000, 1, 0x00000040 },  /* 38 */
{39, "NA-00", "NA-00", {"input", "output", "F5=.DU",     "F4=.DU",       "F0=SPI0_SCLK", "F1=SD3",        "F2=CTS0",    "F3=.RE"           },3, 07000000000, 1, 0x00000080 },  /* 39 */
{40, "NA-00", "NA-00", {"input", "output", "F5=TXD1",    "F4=SPI2_MISO", "F0=PWM0",      "F1=SD4",        "F2=.DU",     "F3=.RE"           },4, 00000000007, 1, 0x00000100 },  /* 40 */
{41, "NA-00", "NA-00", {"input", "output", "F5=RXD1",    "F4=SPI2_MOSI", "F0=PWM1",      "F1=SD5",        "F2=.RE",     "F3=.RE"           },4, 00000000070, 1, 0x00000200 },  /* 41 */
{42, "NA-00", "NA-00", {"input", "output", "F5=RTS1",    "F4=SPI2_SCLK", "F0=GPCLK1",    "F1=SD6",        "F2=.RE",     "F3=.RE"           },4, 00000000700, 1, 0x00000400 },  /* 42 */
{43, "NA-00", "NA-00", {"input", "output", "F5=CTS1",    "F4=SPI2_CE0_N","F0=GPCLK2",    "F1=SD7",        "F2=.RE",     "F3=.RE"           },4, 00000007000, 1, 0x00000800 },  /* 43 */
{44, "NA-00", "NA-00", {"input", "output", "F5=.DU",     "F4=SPI2_CE1_N","F0=GPCLK1",    "F1=SDA0",       "F2=SDA1",    "F3=.RE"           },4, 00000070000, 1, 0x00001000 },  /* 44 */
{45, "NA-00", "NA-00", {"input", "output", "F5=.DU",     "F4=SPI2_CE2_N","F0=PWM1",      "F1=SCL0",       "F2=SCL1",    "F3=.RE"           },4, 00000700000, 1, 0x00002000 },  /* 45 */
{46, "NA-00", "NA-00", {"input", "output", "F5=.DU",     "F4=.DU",       "F0=.RE",       "F1=.DU",        "F2=.DU",     "F3=.DU"           },4, 00007000000, 1, 0x00004000 },  /* 46 */
{47, "NA-00", "NA-00", {"input", "output", "F5=.DU",     "F4=.DU",       "F0=.RE",       "F1=.DU",        "F2=.DU",     "F3=.DU"           },4, 00070000000, 1, 0x00008000 },  /* 47 */
{48, "NA-00", "NA-00", {"input", "output", "F5=.DU",     "F4=.DU",       "F0=.RE",       "F1=.DU",        "F2=.DU",     "F3=.DU"           },4, 00700000000, 1, 0x00010000 },  /* 48 */
{49, "NA-00", "NA-00", {"input", "output", "F5=.DU",     "F4=.DU",       "F0=.RE",       "F1=.DU",        "F2=.DU",     "F3=.DU"           },4, 07000000000, 1, 0x00020000 },  /* 49 */
{50, "NA-00", "NA-00", {"input", "output", "F5=.DU",     "F4=.DU",       "F0=.RE",       "F1=.DU",        "F2=.DU",     "F3=.DU"           },5, 00000000007, 1, 0x00040000 },  /* 50 */
{51, "NA-00", "NA-00", {"input", "output", "F5=.DU",     "F4=.DU",       "F0=.RE",       "F1=.DU",        "F2=.DU",     "F3=.DU"           },5, 00000000070, 1, 0x00080000 },  /* 51 */
{52, "NA-00", "NA-00", {"input", "output", "F5=.DU",     "F4=.DU",       "F0=.RE",       "F1=.DU",        "F2=.DU",     "F3=.DU"           },5, 00000000700, 1, 0x00100000 },  /* 52 */
{53, "NA-00", "NA-00", {"input", "output", "F5=.DU",     "F4=.DU",       "F0=.RE",       "F1=.DU",        "F2=.DU",     "F3=.DU"           },5, 00000007000, 1, 0x00200000 }   /* 53 */
};

int useableAB2[] = 
    {2, 3, 4, 7, 8, 9, 10, 11, 14, 15, 17, 18,
    22, 23, 24, 25, 27, 28, 29, 30, 31};

int useableBplus[] = {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 
    10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 
    20, 21, 22, 23, 24, 25, 26, 27};

/*
 Documentation: register-addresses for 
    Output Set at +1C +20 
    Output Clear +28 +2C
    Input Level at +34 +38
*/

#define SET_GPIO_ALT(g,a) *(gpio+(((g)/10))) |= (((a)<=3?(a)+4:(a)==4?3:2)<<(((g)%10)*3))

#define GPIO_SET *(gpio+7)  // sets   bits which are 1 ignores bits which are 0 register +1C (decimal 28, 7*4)
#define GPIO_CLR *(gpio+10) // clears bits which are 1 ignores bits which are 0 register +28 (decimal 40, 7*10)

char *captxt[] = {"Input", "Output", "F5", "F4", "F0", "F1", "F2", "F3", NULL};

int readcap(volatile uint32_t *gpio, int channel)
{
    uint32_t rg0;
    unsigned c10, c1;

    /* Existing channels are from 0 to 53 */ 
    if(channel < 0 || channel >= 54) return(-1);
    
    c10 = channel / 10;
    c1 = channel % 10;

    rg0 = *(gpio + c10);
    rg0 >>= (3*c1);
    rg0 &= 7;
    return(rg0);
}

int showfsel(volatile uint32_t *gpio, char *gstyl)
{
    int j, caa, gnn;
    int gpiocount;

    if(!strcmp("G26+8", gstyl))
    {
        gpiocount = sizeof(useableAB2) / sizeof(int); 

        for(j = 0; j<gpiocount; j++)
        {
            gnn = useableAB2[j];
            caa = readcap(gpio, gnn);
            printf("GPIO%02u pin %s %s\n", gnn, 
                thegpinfo[gnn].pinnoAB2, thegpinfo[gnn].fx[caa]);
        }
    }
    else
    if(!strcmp("G40", gstyl))
    {
        gpiocount = sizeof(useableBplus) / sizeof(int); 

        for(j = 0; j<gpiocount; j++)
        {
            gnn = useableBplus[j];
            caa = readcap(gpio, gnn);
            printf("GPIO%02u pin %s %s\n", gnn, 
                thegpinfo[gnn].pinnoBplus, thegpinfo[gnn].fx[caa]);
        }
    }
    return(0);
}

/* Look at /proc/cpuinfo and return the revision.
    Revision 0010 is B+, 
    0008 is model A, 
    000e is UK made model B2
    000f is Chinese made model B2
    and so on... 
*/ 
int findrev()
{
    FILE *fp;
    char alin[100];
    char *arr;
    char *kp;
    int fdrv;
    char *stop;

    fdrv = -1;

    fp = fopen("/proc/cpuinfo", "r");
    if(!fp)
    {
        perror("/proc/cpuinfo");
        return(-1);
    }

    for(;;)
    {
        fgets(alin, 100, fp);
        /* fprintf(stderr, "Read %s", alin);  */ 

        if(feof(fp)) break;

        arr = strchr(alin, '\t');
        if(!arr) continue;
        *arr = 0;
        /* fprintf(stderr, "Now >%s< >%s<", alin, arr+1); */ 
        if(strcmp("Revision", alin)) continue;

        kp = strchr(arr+1, ':');
        kp ++;
        fdrv = strtol(kp, &stop, 16); 
        break;  
    }

    fclose(fp);
    return(fdrv);
}

/* Model Pi 2 has a different GPIO base than the earlier models */
uint32_t ModelToGPIOBase(int fdrv)
{
    uint32_t peribase;
    uint32_t gpio0;

    peribase = 0x20000000;

    /* Model Pi 2 (not to be confused with the earlier Pi B revision 2) 
        has revision numbers such as: 0xa21041:
    */ 
    if(fdrv > 0x14)  
    {
        peribase = 0x3F000000;
    }

    gpio0 = peribase + 0x200000;

    return(gpio0);
}


/* Argument: allow us to show build version (date)
    as well as limiting the list of output. */ 
int main(int argc, char **argv)
{
    int mem_fd;
    volatile uint32_t *gpio;
    char *gpio_map;
    int fdrv;
    char gstyl[20];
    char minfo[30];
    uint32_t gpio0;
    int j;
    int devtypeonly;

    devtypeonly = 0;

    /* Look at the arguments. */ 
    for(j = 1; j<argc; j++)
    {
        if(!strcmp("-?", argv[j]))
		{
			printf("usage:\n"
				"gpiostates -v -d\n"
				"where:\n"
				"-v show version\n"
				"-d stop after identifying this machine\n"
				"-? this list\n"
				); 
			return(0);
		}

        if(!strcmp("-v", argv[j]))
        {
            printf("gpiostates %s %s \n", __DATE__, __TIME__); 
        }

        if(!strcmp("-d", argv[j]))
        {
            devtypeonly = 1;
        }
    }

    fdrv = findrev();

    switch(fdrv)
    {
    /* From elinux.org/RPi_HardwareHistory */ 

    case 0x02:
        snprintf(gstyl, 20, "%s", "G26" );
        snprintf(minfo, 30, "%s", "Model B R1 256M" );
        break;

    case 0x03:
        snprintf(gstyl, 20, "%s", "G26" );
        snprintf(minfo, 30, "%s", "Model B R1 256M (ECN0001)" );
        break;

    case 0x04:
        snprintf(gstyl, 20, "%s", "G26+8" );
        snprintf(minfo, 30, "%s", "Model B R2 256M (Sony)" );
        break;
    case 0x05:
        snprintf(gstyl, 20, "%s", "G26+8" );
        snprintf(minfo, 30, "%s", "Model B R2 256M (Qisda)" );
        break;
    case 0x06:
        snprintf(gstyl, 20, "%s", "G26+8" );
        snprintf(minfo, 30, "%s", "Model B R2 256M (Egoman)" );
        break;
    case 0x07:
        snprintf(gstyl, 20, "%s", "G26+8" );
        snprintf(minfo, 30, "%s", "Model A (Egoman)" );
        break;
    case 0x08:
        snprintf(gstyl, 20, "%s", "G26+8" );
        snprintf(minfo, 30, "%s", "Model A (Sony)" );
        break;
    case 0x09:
        snprintf(gstyl, 20, "%s", "G26+8" );
        snprintf(minfo, 30, "%s", "Model A (Qisda)" );
        break;

    case 0x0d:
        snprintf(gstyl, 20, "%s", "G26+8" );
        snprintf(minfo, 30, "%s", "Model B2 512M (Egoman)" );
        break;

    case 0x0e:
        snprintf(gstyl, 20, "%s", "G26+8" );
        snprintf(minfo, 30, "%s", "Model 2 512MB (Sony)" );
        break;

    case 0x0f:
        snprintf(gstyl, 20, "%s", "G26+8" );
        snprintf(minfo, 30, "%s", "Model 2 512MB (Qisda)" );
        break;

    case 0x10:
        snprintf(gstyl, 20, "%s", "G40" );
        snprintf(minfo, 30, "%s", "Model B+" );
        break;

    case 0x11:
        snprintf(gstyl, 20, "%s", "G40" ); // Details TBD! 
        snprintf(minfo, 30, "%s", "Compute Module" );
        break;

    case 0x12:
        snprintf(gstyl, 20, "%s", "G40" );
        snprintf(minfo, 30, "%s", "Model A+" );
        break;

    case 0xa21041:
        snprintf(gstyl, 20, "%s", "G40" );
        snprintf(minfo, 30, "%s", "Model B2 1GB (China)" );
        break;

    case 0xa01041:
        snprintf(gstyl, 20, "%s", "G40" );
        snprintf(minfo, 30, "%s", "Model B2 1GB (UK)" );
        break;

    default:
        snprintf(gstyl, 20, "%s", "G26+8" );
        snprintf(minfo, 30, "%s", "Unrecognized" );
        break;
    }

    printf("Found revision = 0x%x (%s) %s\n", fdrv, gstyl, minfo); 

    if(devtypeonly)
    {
        return(0);
    }

    gpio0 = ModelToGPIOBase(fdrv);

    mem_fd = open("/dev/mem", O_RDWR | O_SYNC);
    if(mem_fd < 0)
    {
        perror("/dev/mem");
        exit(2);
    }
    
    gpio_map = (char *) mmap(NULL, 
        BLOCK_SIZE, 
        PROT_READ|PROT_WRITE,
        MAP_SHARED,
        mem_fd,
        gpio0); 

    close(mem_fd);

    gpio = (volatile uint32_t *) gpio_map;
    showfsel(gpio, gstyl); 
    return(0);
}


