

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <unistd.h>
#include <stdint.h>
#include "devmemgpio.h"


/* Look at /proc/cpuinfo and return the revision.
    Revision 0010 is B+, 
    0008 is model A, 
    000e is UK made model B2
    000f is Chinese made model B2
    and so on... 
*/ 
int findrev()
{
    FILE *fp;
    char alin[100];
    char *arr;
    char *kp;
    int fdrv;
    char *stop;

    fdrv = -1;

    fp = fopen("/proc/cpuinfo", "r");
    if(!fp)
    {
        perror("/proc/cpuinfo");
        return(-1);
    }

    for(;;)
    {
        fgets(alin, 100, fp);
        /* fprintf(stderr, "Read %s", alin);  */ 

        if(feof(fp)) break;

        arr = strchr(alin, '\t');
        if(!arr) continue;
        *arr = 0;
        /* fprintf(stderr, "Now >%s< >%s<", alin, arr+1); */ 
        if(strcmp("Revision", alin)) continue;

        kp = strchr(arr+1, ':');
        kp ++;
        fdrv = strtol(kp, &stop, 16); 
        break;  
    }

    fclose(fp);
    return(fdrv);
}

/* Model Pi 2 has a different periferal base than the earlier models */
uint32_t ModelToPERIBase(int fdrv)
{
    uint32_t peribase;

    peribase = 0x20000000;

    /* Model Pi 2 (not to be confused with the earlier Pi B revision 2) 
        has revision numbers such as: 0xa21041:
    */ 
    if(fdrv > 0x14)  
    {
        peribase = 0x3F000000;
    }
    return(peribase);
}

/* GPIO controller */
uint32_t GPIO_BASE(uint32_t peri_base)  
{ 	
	return(peri_base + 0x200000); 
}  

/* PWM controller */
uint32_t PWM_BASE(uint32_t peri_base)   
{ 
	return(peri_base + 0x20C000); 
}                

/* Clock controller */
uint32_t CLOCK_BASE(uint32_t peri_base) 
{ 
	return(peri_base + 0x101000); 
} 



